rapply <- function(l, handle) {
    if (is.list(l)) {
        handle(lapply(l, function(s) { rapply(s, handle) }))
    } else {
        handle(l)
    }
}

uoptCharState <- function(status) {
    paste(names(status),
          lapply(status, function(t) sprintf(sprintf('%%-%ds', if (is.integer(t)) 4 else 8), as.character(t))),
          collapse=' ', sep='=')
}

uoptShowState <- function(status) {
    cat(uoptCharState(status))
    cat('\n')
}

uopt <- function(x0, f, g, H = NULL, Hv = NULL, Jv = NULL,
                 s = function(x, y, status) uoptShowState(status),
                 options = list(),
                 env = parent.frame()) {
    if (!is.list(options)) {
        stop('options must be a list')
        return();
    }
    if (is.null(options$enableExit)) {
        options$enableExit <- 0
    }
    if (is.null(options$enableSignalHandler)) {
        options$enableSignalHandler <- 0
    }
    if (is.null(options$enableFloatExcept)) {
        options$enableFloatExcept <- 0
    }
    mystatefun <- function(x, y, status) {
        modstatus <- lapply(status, function(x) {
            tryCatch(as.numeric(x), warning=function(cond) x)
        })
        nins <- length(formals(s))
        if (nins == 1) {
            s(modstatus)
        } else if (nins == 2) {
            s(x, modstatus)
        } else {
            s(x, y, modstatus)
        }
    }
    res <- .Call(C_uoptr, x0, f, g, Jv, H, Hv, mystatefun, options, env)
    res
}

showHTMLURL <- function(url) {
    viewer <- getOption("viewer")
    if (!is.null(viewer))
        viewer(url)
    else
        utils::browseURL(url)
}

saveHTMLDoc <- function(doc, tfile=tempfile('rep', fileext='.html')) {
    if ('xml_document' %in% class(doc)) {
        xml2::write_xml(doc, file=tfile)
    } else {
        writeLines(doc, con=tfile)
    }
    tfile
}


showHTMLDoc <- function(doc) {
    tfile <- saveHTMLDoc(doc)
    showHTMLURL(tfile)
    tfile
}

uoptHistSel <- function(hist, name) sapply(hist, '[[', name)

uoptHistData <- function(uopt, select = c())  {
    if (length(select) == 0) {
        select <- names(hist[[1]])
    }
    hist <- uopt$uoptHistory(select)
    l <- lapply(select, function(n) uoptHistSel(hist, n))
    names(l) <- select
    do.call(data.frame, l)
}

uoptReport <- function(uopt, show = TRUE) {

    hist <- uopt$uoptHistory(c('tj', 'it', 'J', 'Er'))

    col <- as.factor(sapply(hist, '[[', 'm'))

    png('dia-st.png')
    plot(sapply(hist, '[[', 'it'), sapply(hist, '[[', 'st'), col=col,
         main = 'Status history', xlab='Iteration', ylab='Status')
    dev.off()

    png('dia-J.png')
    plot(sapply(hist, '[[', 'tj'), sapply(hist, '[[', 'J'), col=col, log="y",
         main = 'Convergence history', xlab='Time (s)', ylab='Objective')
    dev.off()

    png('dia-Er.png')
    plot(sapply(hist, '[[', 'tj'), sapply(hist, '[[', 'Er')/10, col=col, log="y",
         main = 'Relative convergence history', xlab='Time (s)', ylab='Enhancement (%)')
    dev.off()

    if ('cgbeta' %in% names(hist[[1]])) {
        png('dia-cgbeta.png')
        cgbeta <- sapply(hist, '[[', 'cgbeta')
        col <-(cgbeta<0)+1
        col[cgbeta == 0] = 3
        cgbeta[cgbeta == 0] = min(abs(cgbeta[cgbeta != 0]))/2.0
        plot(sapply(hist, '[[', 'tj'), abs(cgbeta), col=col, log="y",
             main = 'CG beta factor', xlab='Time (s)', ylab='beta')
        dev.off()
    }

    cont <- list(
        head = list(
          style = '
             .figure { text-align: center; }
             .figure .caption { font-style: italic; }
             .table .caption { font-style: italic; }
             table { padding: 1em 0em }
             table th { text-align: left;
                        border-bottom: 0.15em solid black;
                        padding: 0.15em 0em 0em 0em }
             table td { padding: 0.15em 0em }
          '
        ),
        report = r2x::element(
            style = 'font-family: sans-serif',
            val = list(
                title = 'UOpt Report',
                section = list(
                    title = 'Summary',
                    table = list(
                        caption = 'Exit status',
                        data =  uopt$uoptHistory(c('exit'))
                    ),
                    table = list(
                        caption = 'User Options',
                        data = uoptOptions(uopt)
                    ),
                    table = list(
                        caption = 'Starting point',
                        data =  uopt$uoptHistory(c('nv'))
                    )
                ),
                section = list(
                    title = 'Convergence History',
                    section = list(
                        title = 'Objective function',
                        para =  'The figure shows the progress of the objective over time.',
                        figure = list(graph = 'dia-J.png', caption = 'J')
                    ),
                    section = list(
                        title = 'Relative Enhancement',
                        para =  'Step-by-step gains in percent over time.',
                        figure = list(graph = 'dia-Er.png', caption = 'E (percent)')
                    ),
                    section = list(
                        cond = function(k) 'cgbeta' %in% names(hist[[1]]),
                        title = 'CG-Beta',
                        para =  'The beta scale factor in CG method.',
                        figure = list(graph = 'dia-cgbeta.png', caption = 'beta')
                    ),
                    section = list(
                        title = 'Status',
                        para =  'The figure shows the optimizer state for each iteration.',
                        figure = list(graph = 'dia-st.png', caption = 'Status')
                    )
                ),
                section = list(
                    title = 'Information',
                    div = list(
                        table = list(
                            caption = 'Software information',
                            data =  uopt$uoptHistory(c('version'))
                        )
                    )
                )
            )
        )
    )

    xsl <- xml2::read_xml(system.file('xsl/uopt-to-html-report.xsl', package = 'uopt'))
    template <- system.file('xml/uopt-report.html', package = 'uopt')

    fcont  <- rapply(cont, function(s) {
        if (is.list(s)) {
            cat(sprintf('Checking section %s\n', s$title))
            if (is.null(s$cond)) {
                cat(sprintf('No condition set in %s\n', s$title))
                s
            } else {
                res <- s$cond()
                cat(sprintf('Condition set in %s: %s, res=%s\n', s$title, paste(deparse(s$cond), collapse=''), res))
                if (res) {
                    s
                } else {
                    NULL
                }
            }
        } else {
            s
        }
    })

    xdoc <- xml2::read_xml(r2x::r2x(fcont))

    recSearchFiles <- function(l) {
        if (is.list(l)) {
            if (!is.null(l$graph))
                l$graph
            else
                lapply(l, recSearchFiles)
        }
    }

    figfiles <- unlist(recSearchFiles(cont))
    names(figfiles) <- rep('data.item', length(figfiles))

    idata <- lapply(figfiles, function(f) {
        file.copy(f, tempdir(), overwrite = TRUE)
        idata <- readBin(con=f, n=file.size(f), 'raw')
        list(id=f, data=base64enc::base64encode(idata, 0))
    })

    writeLines(r2x::r2x(idata), con = (idatadoc <- tempfile('idata')))

    params <- list(`template-doc-uri` = sprintf('file:///%s', URLencode(template)),
                   `image-doc-uri` = sprintf('file:///%s', URLencode(idatadoc)))

    rep <- xslt::xml_xslt(xdoc, xsl, params)

    if (show) {
        repfile <- showHTMLDoc(rep)
    } else {
        repfile <- saveHTMLDoc(rep)
    }

    repfile

}

uoptMovie <- function(uopt, show = TRUE) {

    outfile <- tempfile('mov', fileext = '.mp4')

    r <- system2('ffmpeg', c('-r', '25', '-i', sprintf('%s/state-%%04d.png', tempdir()), '-pix_fmt', 'yuv420p', outfile))

    if (show) {
        showHTMLURL(outfile)
    }

    outfile
}

uoptOptions <- function(uopt, show = TRUE) {

    uopt$uoptOptions()

}

uoptCreate <- function() {
    local({
        hist <- list()

        uoptHistory <- function(select = c()) {
            if (length(select) == 0) {
                hist[TRUE]
            } else {
                hist[sapply(hist, function(t) all(select %in% names(t)))]
            }
        }

        uoptOptions <- function() this$options

        uoptClearHistory <- function() {
            hist <<- list()
        }

        uoptLocal <- function(...) {
            args <- match.call(uopt, as.call(append('uopt', list(...))))
            this$args <<- args
            this$options <<- args$options
            statefun <- args$s
            options <- args$options
            saveplots <- if (!is.null(options$saveplots)) options$saveplots else 0
            if (saveplots) {
                efiles <- dir(tempdir(), 'state-.*.png')
                lapply(efiles, function(f) unlink(sprintf('%s/%s', tempdir(), f)))
            }
            mystatefun <- function(x, status) {
                hist <<- append(hist, list(status))
                if (saveplots) {
                    if (!is.null(status$it)) {
                        png(sprintf('%s/state-%04d.png', tempdir(), status$it))
                        plot(x,
                             xlab='n',
                             ylab='x',
                             main='Current search state',
                             sub=uoptCharState(status[c('J', 'Er', 'it', 'tj', 'st', 'm')]))
                        dev.off()
                    }
                }
                statefun(x, status)
            }
            args$s <- mystatefun
            do.call(uopt, as.list(args)[-1])
        }

        uoptReport <- function(show = TRUE) {
            uoptReport(this, show = show)
        }

        uoptMovie <- function(show = TRUE) {
            uoptMovie(this, show = show)
        }

        this <- list(uopt=uoptLocal,
                     uoptHistory=uoptHistory,
                     uoptClearHistory=uoptClearHistory,
                     uoptReport=uoptReport,
                     uoptMovie=uoptMovie,
                     uoptOptions=uoptOptions)
    })
}
